/*************************************************************************

         name: backgroundKnowledge.pl (Chapter 6)
      version: Dec 10, 2002
  description: Defining Background Knowledge in the form a first-order theory
      authors: Patrick Blackburn & Johan Bos

*************************************************************************/

:- module(backgroundKnowledge,[backgroundKnowledge/2]).

:- use_module(comsemPredicates,[memberList/2,
				basicFormula/1,
				compose/3]).

:- use_module(englishLexicon,[lexEntry/2]).


/*========================================================================
   Dynamic Predicates
========================================================================*/

:- dynamic axiom/3, knowledge/1.


/*========================================================================
   World Knowledge for "Formula" coded as First-order Axioms
========================================================================*/

backgroundKnowledge(Formula,Knowledge):-
   formula2symbols(Formula,[]-Symbols),
   backgroundKnowledge(Formula,Symbols,Knowledge).


/*========================================================================
   World Knowledge coded as First-order Axioms
========================================================================*/

backgroundKnowledge(F1,Symbols1,Knowledge):-
   worldKnowledge(Symbols1,F2),
   formula2symbols(and(F1,F2),[]-Symbols2),
   (
      sort(Symbols1,Sorted), 
      sort(Symbols2,Sorted), 
      !,
      Knowledge=F2
   ;
      backgroundKnowledge(and(F1,F2),Knowledge)
   ).
 

/*========================================================================
   Selecting World Knowledge
========================================================================*/

worldKnowledge(Symbols,Formula):-
   retractall(knowledge(_)),
   worldKnowledge(Symbols),
   knowledge2formula(Formula).

worldKnowledge([]).

worldKnowledge([symbol(Symbol,Arity)|L]):-
   findall(_,(axiom(Symbol,Arity,F),assert(knowledge(F))),_),
   worldKnowledge(L).


/*========================================================================
   Put all selected knowledge in one formula
========================================================================*/

knowledge2formula(F):-
   knowledge(F1),
   retract(knowledge(F1)),
   (
      knowledge(_), !,
      knowledge2formula(F2),
      F=and(F1,F2)
   ; 
      F=F1
   ).
   

/*========================================================================
   World Knowledge
========================================================================*/

%% Only persons or buildings can collapse
%%
axiom(collapse,1,
      all(X,imp(collapse(X),or(person(X),building(X))))).

%% Only persons can dance
%%
axiom(dance,1,all(X,imp(dance(X),person(X)))).

%% Only organisms can die
%%
axiom(die,1,all(X,imp(die(X),organism(X)))).

%% Only animals or persons can growl
%%
axiom(growl,1,all(X,imp(growl(X),or(animal(X),person(X))))).

%% Only artifacts can be cleaned, and only persons can clean
%%
axiom(clean,2,all(X,all(Y,imp(clean(X,Y),and(person(X),artifact(Y)))))).

%% Only beverages can be drunk, and only persons can drink
%%
axiom(drink,2,all(X,all(Y,imp(drink(X,Y),and(person(X),beverage(Y)))))).

%% Only edible food can be eaten, and only persons or animals can eat
%%
axiom(eat,2,all(X,all(Y,imp(eat(X,Y),and(or(person(X),animal(X)),and(edible(Y),food(Y))))))).

%% Nothing can have itself, and no object can be part of two different objects
%%
axiom(have,2,and(not(some(X,have(X,X))),
                 all(X,all(Y,imp(some(Z,and(object(X),and(object(Y),and(object(Z),and(have(X,Z),have(Y,Z)))))),
                                 eq(X,Y)))))).

%% There are at least two different cars
%%
axiom(car,1,some(X,some(Y,and(car(X),and(car(Y),not(eq(X,Y))))))).



/*========================================================================
   Derive all Symbols from a Formula
========================================================================*/

formula2symbols(X,S-S):-
   var(X), !.

formula2symbols(X,S-[symbol(X,0)|S]):-
   atom(X),
   \+ memberList(symbol(X,0),S).

formula2symbols(X,S-S):-
   atom(X),
   memberList(symbol(X,0),S).

formula2symbols(some(_,F),S1-S2):-
   formula2symbols(F,S1-S2).

formula2symbols(all(_,F),S1-S2):-
   formula2symbols(F,S1-S2).

formula2symbols(not(F),S1-S2):-
   formula2symbols(F,S1-S2).

formula2symbols(and(F1,F2),S1-S3):-
   formula2symbols(F1,S1-S2),
   formula2symbols(F2,S2-S3).

formula2symbols(or(F1,F2),S1-S3):-
   formula2symbols(F1,S1-S2),
   formula2symbols(F2,S2-S3).

formula2symbols(imp(F1,F2),S1-S3):-
   formula2symbols(F1,S1-S2),
   formula2symbols(F2,S2-S3).

formula2symbols(F,S1-[symbol(Symbol,1)|S2]):- 
   basicFormula(F),
   compose(F,Symbol,[Arg]),
   \+ memberList(symbol(Symbol,1),S1),
   formula2symbols(Arg,S1-S2).

formula2symbols(F,S1-S2):- 
   basicFormula(F),
   compose(F,Symbol,[Arg]),
   memberList(symbol(Symbol,1),S1),
   formula2symbols(Arg,S1-S2).

formula2symbols(F,S1-[symbol(Symbol,2)|S3]):- 
   basicFormula(F),
   compose(F,Symbol,[Arg1,Arg2]),
   \+ memberList(symbol(Symbol,2),S1),
   formula2symbols(Arg1,S1-S2),
   formula2symbols(Arg2,S2-S3).

formula2symbols(F,S1-S3):- 
   basicFormula(F),
   compose(F,Symbol,[Arg1,Arg2]),
   memberList(symbol(Symbol,2),S1),
   formula2symbols(Arg1,S1-S2),
   formula2symbols(Arg2,S2-S3).


/*========================================================================
   Generate Lexical Knowledge
========================================================================*/

generateLexicalKnowledge:-
   lexEntry(noun,[symbol:Concept,_,hypernyms:Hyp]),
   memberList(Super,Hyp),
   compose(F1,Concept,[X]),
   compose(F2,Super,[X]),
   assert(axiom(Concept,1,all(X,imp(F1,F2)))),
   fail.

generateLexicalKnowledge:-
   lexEntry(pn,[symbol:Sym,_,sem:Hyp]),
   memberList(Concept,Hyp),
   compose(F,Concept,[X]),
   assert(axiom(Sym,0,all(X,imp(eq(X,Sym),F)))),
   fail.

generateLexicalKnowledge:-
   lexEntry(adj,[symbol:Concept1,_,antonyms:Disjoint]),
   memberList(Concept2,Disjoint),
   compose(F1,Concept1,[X]),
   compose(F2,Concept2,[X]),
   assert(axiom(Concept1,1,all(X,imp(F1,not(F2))))),   
   fail.

generateLexicalKnowledge.


/*========================================================================
   Generate lexical knowledge while consulting this file
========================================================================*/

:- generateLexicalKnowledge.
