/*************************************************************************

         name: folResolution.pl (Chapter 5)
      version: Oct 26, 2002
  description: First-Order Resolution Theorem Prover
      authors: Patrick Blackburn & Johan Bos
 
*************************************************************************/

:- module(folResolution,[rprove/1,
                         rprove2/1,
                         rproveTestSuite/0]).

:- use_module(comsemPredicates,[memberList/2,
				appendLists/3,
                                unify/2,
				ordUnion/3,
				selectFromList/3]).

:- use_module(folTestSuite,[formula/2]).

:- use_module(cnfFOL,[cnf/2]).


/*========================================================================
   Main Predicate
========================================================================*/

rprove(Formula):-
   cnf(not(Formula),CNF),
   nonRedundantFactors(CNF,NRF),
   refute(NRF,NRF,Result),
   Result='Theorem'.

rprove2(Formula):-
   cnf(not(Formula),CNF),
   nonRedundantFactors(CNF,NRF),
   refute(NRF,NRF,Result),
   write(Result).


/*========================================================================
   Prove all formulas from the test suite
========================================================================*/

rproveTestSuite:-
   format('~n~n>>>>> FOL RESOLUTION PROVER ON TEST SUITE <<<<<~n',[]),
   formula(Formula,Status),
   format('~n~nInput formula: ~p~nStatus: ~p',[Formula,Status]),
   format('~nProver says: ',[]),
   rprove2(Formula), 
   fail.

rproveTestSuite.


/*========================================================================
   Refute
========================================================================*/

refute(_,Old,'Theorem'):-
   memberList([],Old), !.

refute(Unused,_,'Not a theorem'):-
   Unused = [], !.

refute(Unused,Old,Result):-
   resolveListList(Unused,Old,Old,[],Output),
   ordUnion(Output,Old,NextOld),
   refute(Output,NextOld,Result).


/*========================================================================
   Resolve a list against a list
========================================================================*/

resolveListList([],_,_,Acc,Acc).

resolveListList([H1|T1],[_|T2],Old,Acc,Output):-
   resolveClauseList(H1,T2,Old,Acc,Temp),
   resolveListList(T1,T2,Old,Temp,Output).


/*========================================================================
   Resolve a clause against a list
========================================================================*/

resolveClauseList(_,[],_,Acc,Acc).

resolveClauseList(Clause,[H|L],Old,Acc,Output):-
   resolve(Clause,H,Result0), 
   nonRedundantFactors(Result0,Result),
   !,
   (
       memberList(Result,Old), !,
       resolveClauseList(Clause,L,Old,Acc,Output)
   ;   
       memberList(Result,Acc), !,
       resolveClauseList(Clause,L,Old,Acc,Output)
   ;   
       ordUnion([Result],Acc,NewAcc),
       resolveClauseList(Clause,L,Old,NewAcc,Output)
   ).

resolveClauseList(Clause,[_|L],Old,Acc,Output):-
   resolveClauseList(Clause,L,Old,Acc,Output).


/*========================================================================
   Resolve two clauses
========================================================================*/

resolve(Clause1,Clause2,NewClause):-
   selectFromList(Lit1,Clause1,Temp1),
   selectFromList(not(Lit2),Clause2,Temp2), 
   unify(Lit1,Lit2),
   !, 
   ordUnion(Temp1,Temp2,NewClause).

resolve(Clause1,Clause2,NewClause):-
   selectFromList(not(Lit1),Clause1,Temp1),
   selectFromList(Lit2,Clause2,Temp2),
   unify(Lit1,Lit2),
   ordUnion(Temp1,Temp2,NewClause).


/*========================================================================
   Compute Non-Redundant Factors for a list of clauses
========================================================================*/

nonRedundantFactors([],[]).

nonRedundantFactors([C1|L1],L4):-
   findall(C2,nrFact(C1,C2),L3),
   nonRedundantFactors(L1,L2),
   appendLists(L3,L2,L4).


/*========================================================================
   Compute Non-Redundant Factors for a clause
========================================================================*/

nrFact([],[]).
   
nrFact([X|C1],C2):-
   memberList(X,C1),
   nrFact(C1,C2).

nrFact([X|C1],[X|C2]):-
   nrFact(C1,C2).


/*========================================================================
   Info
========================================================================*/

info:-
   format('~n> -------------------------------------------------------------------- <',[]),
   format('~n> folResolution.pl, by Patrick Blackburn and Johan Bos                 <',[]),
   format('~n>                                                                      <',[]),
   format('~n> ?- rprove(Formula).         - Try to prove Formula                   <',[]),
   format('~n> ?- rprove2(Formula).        - Try to prove Formula, print result     <',[]),
   format('~n> ?- rproveTestSuite.         - run the test suite for theorem proving <',[]),
   format('~n> -------------------------------------------------------------------- <',[]),
   format('~n~n',[]).


/*========================================================================
   Display info at start
========================================================================*/

:- info.

