/*************************************************************************

         name: kellerStorage.pl (Volume 1, Chapter 3)
      version: Oct 16, 2002
  description: Keller Storage Implementation
      authors: Patrick Blackburn & Johan Bos
 
*************************************************************************/

:- module(kellerStorage,[kellerStorage/0,
			 kellerStorage/2,
			 kellerStorageTestSuite/0]).

:- use_module(readLine,[readLine/1]).

:- use_module(comsemPredicates,[memberList/2,
				appendLists/3,
				selectFromList/3,
				compose/3,
				printRepresentations/1]).

:- use_module(alphaConversion,[alphabeticVariants/2]).

:- use_module(betaConversion,[betaConvert/2]).

:- [englishGrammar].

:- use_module(sentenceTestSuite,[sentence/2]).


/*========================================================================
   Driver Predicates
========================================================================*/

kellerStorage:-
   readLine(Sentence),
   setof(Sem,t([sem:Sem],Sentence,[]),Sems1),
   filterAlphabeticVariants(Sems1,Sems2),
   printRepresentations(Sems2).

kellerStorage(Sentence,Sems2):-
   setof(Sem,t([sem:Sem],Sentence,[]),Sems1),
   filterAlphabeticVariants(Sems1,Sems2).


/*========================================================================
   Test Suite Predicates
========================================================================*/

kellerStorageTestSuite:-
   nl, write('>>>>> KELLER STORAGE ON SENTENCE TEST SUITE <<<<< '), nl,
   sentence(Sentence,Readings),
   format('~nSentence: ~p (~p readings)',[Sentence,Readings]),
   kellerStorage(Sentence,Sems),
   printRepresentations(Sems),
   fail.

kellerStorageTestSuite.


/*========================================================================
   Filter Alphabetic Variants
========================================================================*/

filterAlphabeticVariants(L1,L2):-
   selectFromList(X,L1,L3),
   memberList(Y,L3),
   alphabeticVariants(X,Y), !,
   filterAlphabeticVariants(L3,L2).

filterAlphabeticVariants(L,L).


/*========================================================================
   Quantifier Retrieval
========================================================================*/

sRetrieval([S],S).

sRetrieval([Sem|Store],S):-
   selectFromList(bo([Q|NestedStore],X),Store,TempStore),
   appendLists(NestedStore,TempStore,NewStore),
   sRetrieval([app(Q,lam(X,Sem))|NewStore],S).


/*========================================================================
   Composition Rules
========================================================================*/

combine(t:Converted,[s:Sem]):- 
   betaConvert(Sem,Converted).

combine(t:Converted,[q:Sem]):- 
   betaConvert(Sem,Converted).

combine(s:app(A,B),[s:A,s:B]).
combine(s:lam(B,imp(S,B)),[if:S]).
combine(s:lam(B,or(S,B)),[either:S]).
combine(s:S,[then:S]).
combine(s:S,[or:S]).
combine(s:S,[np:[A|S1],vp:[B|S2]]):-
   appendLists(S1,S2,S3),
   sRetrieval([app(A,B)|S3],Retrieved),
   betaConvert(Retrieved,S).

combine(q:Q,[whnp:[A|S1],vp:[B|S2]]):-
   sRetrieval([app(A,B)|S2],VP),
   sRetrieval([VP|S1],Retrieved),
   betaConvert(Retrieved,Q).

combine(np:A,[pn:A]).
combine(np:A,[qnp:A]).
combine(np:[lam(P,app(P,X)),bo([app(A,B)|S],X)],[det:[A],n:[B|S]]).
combine(np:[app(A,B)|S],[det:[A],n:[B|S]]).
combine(np:[app(app(B,A),C)|S3],[np:[A|S1],coord:[B],np:[C|S2]]):-
   appendLists(S1,S2,S3).

combine(whnp:[lam(P,app(P,X)),bo([app(A,B)|S],X)],[det:[A],n:[B|S]]).
combine(whnp:[lam(P,app(P,X)),bo([A],X)],[qnp:[A]]).

combine(n:[app(A,B)|S],[adj:[A],n:[B|S]]).
combine(n:A,[noun:A]).
combine(n:[app(B,A)|S],[noun:[A],nmod:[B|S]]).
combine(n:[app(app(B,A),C)|S3],[n:[A|S1],coord:[B],n:[C|S2]]):-
	appendLists(S1,S2,S3).

combine(nmod:A,[pp:A]).
combine(nmod:A,[rc:A]).
combine(nmod:[lam(P,app(A,app(B,P)))|S3],[pp:[A|S1],nmod:[B|S2]]):-
   appendLists(S1,S2,S3).

combine(vp:[app(A,B)|S],[av:[A],vp:[B|S]]).
combine(vp:[app(A,B)|S],[cop:[A],np:[B|S]]).
combine(vp:A,[iv:A]).
combine(vp:[app(A,B)|S],[tv:[A],np:[B|S]]).
combine(vp:[app(app(B,A),C)|S3],[vp:[A|S1],coord:[B],vp:[C|S2]]):-
	appendLists(S1,S2,S3).

combine(pp:[app(A,B)|S],[prep:[A],np:[B|S]]).

combine(rc:[app(A,B)|S],[relpro:[A],vp:[B|S]]).



/*========================================================================
   Semantic Macros
========================================================================*/

semMacro(det,M):-
   M = [type:uni,
        sem:[lam(P,lam(Q,all(X,imp(app(P,X),app(Q,X)))))]].

semMacro(det,M):-
   M = [type:indef,
        sem:[lam(P,lam(Q,some(X,and(app(P,X),app(Q,X)))))]].

semMacro(det,M):-
   M = [type:wh,
        sem:[lam(P,lam(Q,que(X,app(P,X),app(Q,X))))]].

semMacro(pn,M):-
   M = [symbol:Sym,
        sem:[lam(P,app(P,Sym))]].

semMacro(noun,M):-
   M = [symbol:Sym,
        sem:[lam(X,Formula)]],
   compose(Formula,Sym,[X]).

semMacro(iv,M):-
   M = [symbol:Sym,
        sem:[lam(X,Formula)]],
   compose(Formula,Sym,[X]).

semMacro(tv,M):-
   M = [symbol:Sym,
        sem:[lam(K,lam(Y,app(K,lam(X,Formula))))]], 
   compose(Formula,Sym,[Y,X]).

semMacro(qnp,M):-
   M = [type:wh,
        symbol:Sym,
        sem:[lam(Q,que(X,Formula,app(Q,X)))]], 
   compose(Formula,Sym,[X]).

semMacro(cop,M):-
   M = [pol:pos,
        sem:[lam(K,lam(Y,app(K,lam(X,eq(Y,X)))))]];
   M = [pol:neg,
        sem:[lam(K,lam(Y,not(app(K,lam(X,eq(Y,X))))))]].

semMacro(relpro,M):-
   M = [sem:[lam(P,lam(Q,lam(X,and(app(P,X),app(Q,X)))))]].

semMacro(prep,M):-
   M = [symbol:Sym,
        sem:[lam(K,lam(P,lam(Y,and(app(K,lam(X,F)),app(P,Y)))))]],
   compose(F,Sym,[Y,X]).

semMacro(adj,M):-
   M = [symbol:Sym,
        sem:[lam(P,lam(X,and(F,app(P,X))))]],
   compose(F,Sym,[X]).

semMacro(av,M):-
   M = [pol:neg,
        sem:[lam(P,lam(X,not(app(P,X))))]];
   M = [pol:pos,
        sem:[lam(P,lam(X,app(P,X)))]].

semMacro(coord,M):-
   M = [type:conj,
        sem:[lam(X,lam(Y,lam(P,and(app(X,P),app(Y,P)))))]];  
   M = [type:disj,
        sem:[lam(X,lam(Y,lam(P,or(app(X,P),app(Y,P)))))]].


/*========================================================================
   Info
========================================================================*/

info:-
   format('~n> ------------------------------------------------------------------- <',[]),
   format('~n> kellerStorage.pl, by Patrick Blackburn and Johan Bos                <',[]),
   format('~n>                                                                     <',[]),
   format('~n> ?- kellerStorage.            - parse a typed-in sentence            <',[]),
   format('~n> ?- kellerStorage(S,F).       - parse a sentence and return formulas <',[]),
   format('~n> ?- kellerStorageTestSuite.   - run the test suite                   <',[]),
   format('~n> ------------------------------------------------------------------- <',[]),
   format('~n~n',[]).


/*========================================================================
   Display info at start
========================================================================*/

:- info.
