/*************************************************************************

         name: lambda.pl (Volume 1, Chapter 2)
      version: Oct 16, 2002
  description: Semantic Construction with Beta Conversion
      authors: Patrick Blackburn & Johan Bos
 
*************************************************************************/

:- module(lambda,[lambda/0,lambda/2,lambdaTestSuite/0]).

:- use_module(readLine,[readLine/1]).

:- use_module(comsemPredicates,[printRepresentations/1,
				compose/3]).

:- use_module(betaConversion,[betaConvert/2]).

:- [englishGrammar].

:- use_module(sentenceTestSuite,[sentence/2]).


/*========================================================================
   Driver Predicates
========================================================================*/

lambda:-
	readLine(Sentence),
        lambda(Sentence,Sems),
	printRepresentations(Sems).

lambda(Sentence,Sems):-
	setof(Sem,t([sem:Sem],Sentence,[]),Sems).


/*========================================================================
   Test Suite Predicates
========================================================================*/

lambdaTestSuite:-
	nl, write('>>>>> LAMBDA ON SENTENCE TEST SUITE <<<<< '), nl,
        sentence(Sentence,_),
        nl, write('Sentence: '), write(Sentence),
	lambda(Sentence,Formulas),
	printRepresentations(Formulas),
        fail.

lambdaTestSuite.


/*========================================================================
   Parsing 
========================================================================*/

parse(Text,Formula):-
	t([sem:Formula],Text,[]).


/*========================================================================
   Composition Rules
========================================================================*/

combine(t:Converted,[s:Sem]):- 
   betaConvert(Sem,Converted).

combine(s:app(A,B),[np:A,vp:B]).
combine(s:app(A,B),[s:A,s:B]).
combine(s:lam(B,imp(S,B)),[if:S]).
combine(s:lam(B,or(S,B)),[either:S]).
combine(s:S,[then:S]).
combine(s:S,[or:S]).

combine(np:app(app(B,A),C),[np:A,coord:B,np:C]).
combine(np:app(A,B),[det:A,n:B]).
combine(np:A,[pn:A]).
combine(np:A,[qnp:A]).

combine(n:app(app(B,A),C),[n:A,coord:B,n:C]).
combine(n:app(A,B),[adj:A,n:B]).
combine(n:A,[noun:A]).
combine(n:app(B,A),[noun:A,nmod:B]).

combine(nmod:A,[pp:A]).
combine(nmod:A,[rc:A]).
combine(nmod:lam(P,app(A,app(B,P))),[pp:A,nmod:B]).

combine(vp:app(app(B,A),C),[vp:A,coord:B,vp:C]).
combine(vp:app(A,B),[av:A,vp:B]).
combine(vp:app(A,B),[cop:A,np:B]).
combine(vp:A,[iv:A]).
combine(vp:app(A,B),[tv:A,np:B]).

combine(pp:app(A,B),[prep:A,np:B]).

combine(rc:app(A,B),[relpro:A,vp:B]).


/*========================================================================
   Semantic Macros
========================================================================*/

semMacro(det,M):-
   M = [type:uni,
        sem:lam(U,lam(V,all(X,imp(app(U,X),app(V,X)))))].

semMacro(det,M):-
   M = [type:indef,
        sem:lam(P,lam(Q,some(X,and(app(P,X),app(Q,X)))))].

semMacro(det,M):-
   M = [type:wh,
        sem:lam(P,lam(Q,que(X,app(P,X),app(Q,X))))].

semMacro(pn,M):-
   M = [symbol:Sym,
        sem:lam(P,app(P,Sym))].

semMacro(noun,M):-
   M = [symbol:Sym,
        sem:lam(X,Formula)],
   compose(Formula,Sym,[X]).

semMacro(iv,M):-
   M = [symbol:Sym,
        sem:lam(X,Formula)],
   compose(Formula,Sym,[X]).

semMacro(tv,M):-
   M = [symbol:Sym,
        sem:lam(K,lam(Y,app(K,lam(X,Formula))))], 
   compose(Formula,Sym,[Y,X]).

semMacro(qnp,M):-
   M = [type:wh,
        symbol:Sym,
        sem:lam(Q,que(X,Formula,app(Q,X)))], 
   compose(Formula,Sym,[X]).

semMacro(cop,M):-
   M = [pol:pos,
        sem:lam(K,lam(Y,app(K,lam(X,eq(Y,X)))))];
%   M = [pol:neg,
%        sem:lam(K,lam(Y,app(K,lam(X,not(eq(Y,X))))))].
   M = [pol:neg,
        sem:lam(K,lam(Y,not(app(K,lam(X,eq(Y,X))))))].

semMacro(relpro,M):-
   M = [sem:lam(P,lam(Q,lam(X,and(app(P,X),app(Q,X)))))].

semMacro(prep,M):-
   M = [symbol:Sym,
        sem:lam(K,lam(P,lam(Y,and(app(K,lam(X,F)),app(P,Y)))))],
   compose(F,Sym,[Y,X]).

semMacro(adj,M):-
   M = [symbol:Sym,
        sem:lam(P,lam(X,and(F,app(P,X))))],
   compose(F,Sym,[X]).

semMacro(av,M):-
   M = [pol:neg,
        sem:lam(P,lam(X,not(app(P,X))))];
   M = [pol:pos,
        sem:lam(P,lam(X,app(P,X)))].

semMacro(coord,M):-
   M = [type:conj,
        sem:lam(X,lam(Y,lam(P,and(app(X,P),app(Y,P)))))];  
   M = [type:disj,
        sem:lam(X,lam(Y,lam(P,or(app(X,P),app(Y,P)))))].



/*========================================================================
   Info
========================================================================*/

info:-
   format('~n> ------------------------------------------------------------------ <',[]),
   format('~n> lambda.pl, by Patrick Blackburn and Johan Bos                      <',[]),
   format('~n>                                                                    <',[]),
   format('~n> ?- lambda.              - parse a typed-in sentence                <',[]),
   format('~n> ?- lambda(S,F).         - parse a sentence and return formula      <',[]),
   format('~n> ?- lambdaTestSuite.     - run the test suite                       <',[]),
   format('~n> ------------------------------------------------------------------ <',[]),
   format('~n~n',[]).


/*========================================================================
   Display info at start
========================================================================*/

:- info.
