/*************************************************************************

         name: pluggingAlgorithm.pl
      version: Oct 17, 2002
  description: Plugging Algorithm in Prolog
       author: Johan Bos
 
*************************************************************************/

/*========================================================================
     Module Declaration
========================================================================*/

:- module(pluggingAlgorithm,[plugUF/2]).

:- use_module(comsemPredicates,[compose/3,
                                memberList/2,
                                selectFromList/3]).


/*========================================================================
     Declaration of dynamic predicates
========================================================================*/

:- dynamic plug/2, leq/2, hole/1, label/1.
:- dynamic some/3, all/3, que/4.
:- dynamic not/2, or/3, imp/3, and/3.
:- dynamic pred1/3, pred2/4, eq/3.


/*========================================================================
   Main Plugging Predicate (all pluggings)
========================================================================*/

plugUF(UF,Sem):-
   numbervars(UF,0,_),  % assuming distinct variables
   initUF,
   assertUF(UF),
   top(Top),  
   findall(H,hole(H),Holes),
   findall(L,(label(L),\+ parent(_,L)),Labels),
   plugHoles(Holes,Labels,[]),
   uf2fol(Top,Sem).


/*========================================================================
   Asserting immediate dominance relations to the Prolog database
========================================================================*/
    
parent(A,B):- imp(A,B,_).
parent(A,B):- imp(A,_,B).
parent(A,B):- or(A,B,_).
parent(A,B):- or(A,_,B).
parent(A,B):- and(A,B,_).
parent(A,B):- and(A,_,B).
parent(A,B):- not(A,B).
parent(A,B):- all(A,_,B).
parent(A,B):- some(A,_,B).
parent(A,B):- que(A,_,B,_).
parent(A,B):- que(A,_,_,B).
parent(A,B):- plug(A,B).


/*========================================================================
   Transitive Closure of Dominance
========================================================================*/

dom(X,Y):- dom([],X,Y).

dom(L,X,Y):- 
   parent(X,Y), 
   \+ memberList(parent(X,Y),L).

dom(L,X,Y):- 
   leq(Y,X), 
   \+ memberList(leq(Y,X),L).

dom(L,X,Z):- 
   parent(X,Y), 
   \+ memberList(parent(X,Y),L),
   dom([parent(X,Y)|L],Y,Z).

dom(L,X,Z):- 
   leq(Y,X), 
   \+ memberList(leq(Y,X),L),
   dom([leq(Y,X)|L],Y,Z).


/*========================================================================
   Plugging Holes with Labels
========================================================================*/

plugHoles([],_,Plugs):-
   properPlugging(Plugs).

plugHoles([H|Holes],Labels1,Plugs):-
   properPlugging(Plugs),
   selectFromList(L,Labels1,Labels2),
   plugHoles(Holes,Labels2,[plug(H,L)|Plugs]).


/*========================================================================
   Check whether plugging is propers
========================================================================*/

properPlugging(Plugs):-
   retractall(plug(_,_)), 
   findall(X,(memberList(X,Plugs),assert(X)),_),
   \+ dom(A,A),
   \+ ( parent(A,B), parent(A,C), \+ B=C, dom(B,D), dom(C,D)).
   
 
/*========================================================================
    Top of a UF
========================================================================*/

top(X):- dom(X,_), \+ dom(_,X), !.


/*========================================================================
   From UFs to FOLs
========================================================================*/

uf2fol(H,F):-
   hole(H),
   plug(H,L),
   uf2fol(L,F).

uf2fol(L,all(X,F)):-
   all(L,X,H),
   uf2fol(H,F).

uf2fol(L,some(X,F)):-
   some(L,X,H),
   uf2fol(H,F).

uf2fol(L,que(X,F1,F2)):-
   que(L,X,H1,H2),
   uf2fol(H1,F1),
   uf2fol(H2,F2).

uf2fol(L,imp(F1,F2)):-
   imp(L,H1,H2),
   uf2fol(H1,F1),
   uf2fol(H2,F2).

uf2fol(L,and(F1,F2)):-
   and(L,H1,H2),
   uf2fol(H1,F1),
   uf2fol(H2,F2).

uf2fol(L,or(F1,F2)):-
   or(L,H1,H2),
   uf2fol(H1,F1),
   uf2fol(H2,F2).

uf2fol(L,not(F)):-
   not(L,H),
   uf2fol(H,F).

uf2fol(L,eq(X,Y)):-
   eq(L,X,Y).

uf2fol(L,F):-
   pred1(L,Symbol,Arg),
   compose(F,Symbol,[Arg]).

uf2fol(L,F):-
   pred2(L,Symbol,Arg1,Arg2),
   compose(F,Symbol,[Arg1,Arg2]).


/*========================================================================
   Assert UF to Prolog database
========================================================================*/

assertUF(some(_,F)):-   
    assertUF(F).

assertUF(and(F1,F2)):-
    assertUF(F1),
    assertUF(F2).

assertUF(F):- 
    \+ F=and(_,_),
    \+ F=some(_,_),
    assert(F).


/*========================================================================
   Initialisation 
========================================================================*/

initUF:-
   retractall(hole(_)), retractall(label(_)), retractall(leq(_,_)), 
   retractall(some(_,_,_)), retractall(all(_,_,_)), retractall(que(_,_,_,_)), 
   retractall(pred1(_,_,_)), retractall(pred2(_,_,_,_)), retractall(eq(_,_,_)), 
   retractall(and(_,_,_)), retractall(or(_,_,_)), retractall(not(_,_)), 
   retractall(plug(_,_)), retractall(imp(_,_,_)).

